import sys
from inspect import Signature
from types import CodeType, FunctionType
from typing import Any, Tuple


if sys.version_info >= (3, 8):
    copy_code = CodeType.replace
else:
    PY_36_37_CODE_ARGS: Tuple[str, ...] = (
        "co_argcount",
        "co_kwonlyargcount",
        "co_nlocals",
        "co_stacksize",
        "co_flags",
        "co_code",
        "co_consts",
        "co_names",
        "co_varnames",
        "co_filename",
        "co_name",
        "co_firstlineno",
        "co_lnotab",
        "co_freevars",
        "co_cellvars",
    )

    def copy_code(code: CodeType, **update: Any) -> CodeType:
        """
        Create a copy of code object with changed attributes
        """
        new_args = [update.pop(arg, getattr(code, arg)) for arg in PY_36_37_CODE_ARGS]
        if update:
            raise TypeError(f"Unexpected code attribute(s): {update}")
        return CodeType(*new_args)


def copy_func(f: FunctionType, name, defaults, signature: Signature):
    """
    Makes exact copy of a function object with given name and defaults
    """
    new_defaults = []
    kw_only_defaults = f.__kwdefaults__.copy() if f.__kwdefaults__ else {}

    for key, param in signature.parameters.items():
        if param.kind is param.KEYWORD_ONLY:
            if key in defaults:
                kw_only_defaults[key] = defaults.pop(key)
        elif key in defaults:
            new_defaults.append(defaults.pop(key))
        elif param.default is not param.empty:
            new_defaults.append(param.default)

    new_func = FunctionType(
        code=copy_code(f.__code__, co_name=name),
        globals=f.__globals__,
        name=name,
        argdefs=tuple(new_defaults),
        closure=f.__closure__,
    )
    new_func.__kwdefaults__ = kw_only_defaults
    new_func.__dict__.update(f.__dict__)
    return new_func



import inspect
import itertools
import sys
from contextlib import suppress
from functools import partial, wraps
from types import FrameType, FunctionType, ModuleType
from typing import Any, Dict, Iterable, List, Set, Tuple, Union, cast

from parametrize.utils import copy_func


Parameter = Tuple[str, Any]
Parameters = Tuple[Parameter, ...]
ParametersList = List[Parameters]


class UnparametrizedMethod:
    __slots__ = ("func",)

    def __init__(self, func: FunctionType):
        self.func = func

    def __getattribute__(self, item):
        """
        Forbid any usage of unparametrized object
        """
        if item in {"func", "__repr__", "__name__"}:
            return object.__getattribute__(self, item)

        raise AttributeError("UnparametrizedMethod should not be used anywhere")

    def __repr__(self):
        return f"{self.func.__name__}[...]"


class ParametrizeContext:
    __slots__ = (
        "func",
        "parametrizes_left",
        "all_parameters",
        "seen_argnames",
        "signature",
        "decoration_frame",
    )

    def __init__(self, func: FunctionType, decoration_frame: FrameType):
        self.func = func
        self.signature = inspect.signature(func)
        self.parametrizes_left = _count_parametrize_decorators(func, decoration_frame)
        self.all_parameters: List[ParametersList] = []
        self.seen_argnames: Set[str] = set()
        self.decoration_frame = decoration_frame

    def add(self, parameters: ParametersList, argnames_set: Set[str]):
        reused_names = argnames_set & self.seen_argnames
        if reused_names:
            raise TypeError(f"Arguments names reused: {reused_names}")

        if argnames_set - self.signature.parameters.keys():
            raise TypeError(
                f"Unexpected argument(s) {argnames_set} "
                f"for function {self.func.__name__}{self.signature}"
            )

        self.all_parameters.append(parameters)
        self.seen_argnames.update(argnames_set)
        self.parametrizes_left -= 1

    @property
    def combined_parameters(self):
        for case in itertools.product(*self.all_parameters):
            yield {k: v for params in case for k, v in params}

    def __call__(self, *args, **kwargs):
        """
        We should never end up here.
        This method is necessary just to ensure in case something goes wrong, it won't be unnoticed
        """
        raise RuntimeError(
            f"Attempt to execute {self.func.__qualname__} before it was parametrized"
        )


def parametrize(argnames: Union[str, Iterable[str]], argvalues: Iterable[Any]):
    """
    class TestSomething(unittest.TestCase):

        @parametrize('a,b', [(1, 2), (3, 4)])
        def test_a_and_b(self, a, b):
            self.assertGreater(b, a)

    Trick to use pytest.mark.parametrize with unittest.TestCase

    It generates parametrized test cases and injects them into class namespace
    """

    parameters, argnames_set = _collect_parameters(argnames, argvalues)

    def decorator(
        func_or_context: Union[FunctionType, ParametrizeContext]
    ) -> Union[ParametrizeContext, UnparametrizedMethod]:
        if isinstance(func_or_context, UnparametrizedMethod):  # we should never end up here
            raise RuntimeError(
                "Failed to complete parametrization. "
                "Please make sure all parametrization done with decorators grouped in once place"
            )

        if isinstance(func_or_context, ParametrizeContext):
            context = func_or_context
        else:
            decoration_frame = cast(FrameType, inspect.currentframe().f_back)  # type: ignore
            context = ParametrizeContext(func_or_context, decoration_frame)

        context.add(parameters, argnames_set)

        if context.parametrizes_left:
            return context  # pass context to the next parametrize decorator
        else:
            # set parametrized functions in place of given one
            _set_test_cases(context)
            return UnparametrizedMethod(context.func)

    decorator.__parametrize_decorator__ = parametrize  # type: ignore

    return decorator


def _collect_parameters(argnames, argvalues) -> Tuple[ParametersList, Set[str]]:
    if isinstance(argnames, str):
        argnames = list(map(str.strip, argnames.split(",")))

    argnames_set = set(argnames)

    if len(argnames) != len(argnames_set):
        raise TypeError("Arguments must not repeat")

    parameters = []
    for i, values in enumerate(argvalues):
        if len(argnames) == 1 and isinstance(values, str) or not isinstance(values, Iterable):
            values = (values,)

        if len(values) != len(argnames):
            raise ValueError(
                f"Wrong number of values at index {i}, expected "
                f"{len(argnames)}, got {len(values)}: {values}"
            )

        parameters.append(tuple(zip(argnames, values)))

    return parameters, argnames_set


def _find_possible_decorators(
    namespace: Dict[str, Any], search_in_modules: bool = True
) -> Set[str]:
    possible_definitions: Set[str] = set()

    if not search_in_modules and namespace.get("__name__") in sys.builtin_module_names:
        # don't search in builtin modules
        return possible_definitions

    for key, value in namespace.items():
        with suppress(ValueError):  # inspect.unwrap() may raise ValueError
            if (
                value is parametrize
                or getattr(value, "__parametrize_decorator__", False) is parametrize
                or (
                    hasattr(value, "__dict__")
                    and "__wrapped__" in value.__dict__
                    and inspect.unwrap(value) is parametrize
                )
            ):
                possible_definitions.add(key)
            elif search_in_modules and isinstance(value, ModuleType):
                # allow usages like @my_module.my_predefined_params
                possible_definitions.update(
                    _find_possible_decorators(value.__dict__, search_in_modules=False)
                )

    return possible_definitions


def _count_parametrize_decorators(function, decoration_frame):
    possible_definitions = _find_possible_decorators(
        {**decoration_frame.f_globals, **decoration_frame.f_locals}
    )
    lines, _ = inspect.getsourcelines(function)

    # maybe it would be safer/better to use ast.parse for that
    # but for now this method works pretty well
    parametrized_count = 0
    parametrize_decorators_should_end = False
    decorator_out_of_order = False
    for line in map(str.strip, lines):
        if line.startswith("def "):
            break
        for definition in possible_definitions:
            if line.startswith(f"@{definition}"):
                if parametrize_decorators_should_end:
                    raise TypeError(
                        f"Parametrize decorator {line} must be grouped "
                        f"together with other parametrize decorators"
                    )
                parametrized_count += 1
                break
        else:
            is_another_decorator = line.startswith("@")
            if is_another_decorator and parametrized_count:
                parametrize_decorators_should_end = True
            elif is_another_decorator:
                decorator_out_of_order = line

    if not parametrized_count:
        raise RuntimeError(
            f"Unable to find any parametrizes in decorators, "
            f"please rewrite decorator name to match any of detected names @{possible_definitions}"
        )

    if decorator_out_of_order:
        raise TypeError(
            f"{decorator_out_of_order} must be defined before any of "
            f"@{possible_definitions} decorators"
        )

    return parametrized_count


def _set_test_cases(context):
    func = context.func
    used_names: Set[str] = set()
    namespace = context.decoration_frame.f_locals

    for params in context.combined_parameters:
        parameters_str = "-".join(str(v).replace(".", "-") for v in params.values())

        methods_with_same_name = 1
        final_parameters_str = parameters_str

        while final_parameters_str in used_names:
            final_parameters_str = f"{parameters_str}:{methods_with_same_name}"
            methods_with_same_name += 1

        func_name = func.__name__
        used_names.add(final_parameters_str)
        parametrized_name = f"{func_name}[{final_parameters_str}]"
        if parametrized_name in namespace:
            raise NameError(
                f"{func_name!r} parametrized with [{final_parameters_str}] is already defined above"
            )

        # creating a wrapper function.
        # functools.partial alone will not bound to class
        # functools.partialmethod and other descriptors won't be detected as tests
        def get_parametrized_method(f, name, parameters):
            parametrized_func = partial(f, **parameters)

            # copying func with new default parameters and name is necessary for introspection
            # without it, pytest, for example would think that parametrized values are fixtures
            @wraps(copy_func(f, name, parameters, context.signature))
            def parametrized_method(*args, **kwargs):
                return parametrized_func(*args, **kwargs)

            return parametrized_method

        namespace[parametrized_name] = get_parametrized_method(func, parametrized_name, params)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parametrize']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'parametrize',
    'version': '0.1.1',
    'description': 'Drop-in @pytest.mark.parametrize replacement working with unittest.TestCase',
    'long_description': '[![CI](https://github.com/MrMrRobat/parametrize/workflows/CI/badge.svg?event=push)](https://github.com/MrMrRobat/parametrize/actions?query=event%3Apush+branch%3Amaster+workflow%3ACI)\n[![PyPi](https://img.shields.io/pypi/v/parametrize.svg)](https://pypi.python.org/pypi/parametrize)\n[![Python Versions](https://img.shields.io/pypi/pyversions/parametrize.svg)](https://github.com/MrMrRobat/parametrize)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## Drop-in `@pytest.mark.parametrize` replacement working with `unittest.TestCase`\n\n### Why?\nYou want to start using `@pytest.mark.parametrize`, but can\'t simply drop `unittest.TestCase` because you have tons of `self.assert`\'s, `setUp`\'s `tearDown`\'s to rewrite?\n\nWith `@parametrize` you can start parameterizing your tests now, and get rid of `unittest.TestCase` later if needed.\n\n## Usage\n### Simple example from [pytest docs](https://docs.pytest.org/en/6.2.x/parametrize.html) adapted to unittest\n```python\nimport unittest\nfrom parametrize import parametrize\n\nclass TestSomething(unittest.TestCase):\n\n    @parametrize(\'test_input,expected\', [("3+5", 8), ("2+4", 6)])\n    def test_eval(self, test_input, expected):\n        self.assertEqual(expected, eval(test_input))\n```\n```py\n$ python -m unittest test.py -v\ntest_eval[2+4-6] (test.TestSomething) ... ok\ntest_eval[3+5-8] (test.TestSomething) ... ok\ntest_eval[6*9-42] (test.TestSomething) ... FAIL\n\n======================================================================\nFAIL: test_eval[6*9-42] (test.TestSomething)\n----------------------------------------------------------------------\nTraceback (most recent call last):\n  File "parametrize/parametrize.py", line 261, in parametrized_method\n    return parametrized_func(*args, **kwargs)\n  File "test.py", line 8, in test_eval\n    self.assertEqual(expected, eval(test_input))\nAssertionError: 42 != 54\n\n----------------------------------------------------------------------\nRan 3 tests in 0.001s\n\nFAILED (failures=1)\n```\n##### You don\'t need to use additional decorators, custom base classes or metaclasses.\n\n### Stacking parametrize decorators:\n```python\nimport unittest\nfrom parametrize import parametrize\n\nclass TestSomething(unittest.TestCase):\n    \n    @parametrize("x", [0, 1])\n    @parametrize("y", [2, 3])\n    def test_foo(self, x, y):\n        pass\n```\n`test_foo` will be called with: `(x=0, y=2)`, `(x=1, y=2)`, `(x=0, y=3)`, and `(x=1, y=3)`:\n```python\n$ python -m unittest test.py -v\ntest_foo[2-0] (test.TestSomething) ... ok\ntest_foo[2-1] (test.TestSomething) ... ok\ntest_foo[3-0] (test.TestSomething) ... ok\ntest_foo[3-1] (test.TestSomething) ... ok\n\n----------------------------------------------------------------------\nRan 4 tests in 0.000s\n\nOK\n```\n##### Note: even though the tests are always generated in the same order, the execution order is not guaranteed\n\n\n## Compatibility \nAny `@parametrize` decorator can be converted to `@pytest.mark.parametrize` just by changing its name. \n`@pytest.mark.parametrize` decorator can be converted to `@parametrize` as long as `pytest.param`, `indirect`, `ids` and `scope` are not used.\n\n`@parametrize` works with both `unittest` and `pytest`. However, `pytest` is recommended due to [limitations when using unittest in cli](#parametrized-method-can-be-ran-from-command-line-only-via-pytest). \n\nParametrized tests are properly detected and handled by PyCharm. They are displayed as if they were parametrized with `@pytest.mark.parametrize`.\n\n\n## Limitations\nSince `@parametrize` does some kind of magic under the hood, there are some limitations you need to consider.\nIt\'s likely you will never face most of them, but if you will, `@parametrize` will let you know with an error:\n\n- ### All parametrization must be done via decorators\n    :white_check_mark: OK\n    ```python\n    @parametrize(\'a\', (1, 2))\n    def f(a):\n        ...\n    ```   \n    :x: Won\'t work:\n    ```python\n    def f(a):\n        ...\n  \n    parametrize(\'a\', (1, 2))(func)\n    ```\n    ```py\n    RuntimeError: Unable to find any parametrizes in decorators, please rewrite decorator name to match any of detected names @{\'parametrize\'}\n    ```\n    \n- ### All other decorators must be defined before parametrize decorators\n    :white_check_mark: OK:\n    ```py\n    @parametrize("a", (1, 2))\n    @parametrize("b", (2, 3))\n    @mock.patch(f"{__name__}.bar", "foo")\n    def f(a, b):\n        return a, b\n    ```\n    :x: Won\'t work:\n    ```python\n    @mock.patch(f"{__name__}.bar", "foo")\n    @parametrize("a", (1, 2))\n    @parametrize("b", (2, 3))\n    def f(a, b):\n        return a, b\n    ```\n    ```py\n    TypeError: @mock.patch(f"{__name__}.bar", "foo") must be defined before any of @{\'parametrize\'} decorators\n    ```\n\n- ### If you assign parametrized decorator to variable, it must be accessible from `locals()` or `globals()` namespaces:\n    :white_check_mark: OK:\n    ```py\n    a_parameters = parametrize("a", (4, 5))  # defined in module\n    def func():  \n        class TestSomething:\n            b_parameters = parametrize("b", (1, 2, 3))\n  \n            @b_parameters  # b_parameters found in locals()\n            @a_parameters  # a_parameters found in globals()\n            def test_method(self, a, b):\n                ...\n    ```\n    :x: Won\'t work:\n    ```py\n    def func():\n        # defined in function scope\n        a_parameters = parametrize("a", (4, 5))\n    \n        class TestSomething:\n            print(\'a_parameters\' in {**globals(), **locals()})  # False\n    \n            @a_parameters  # accessed in class body scope\n            def test_method(self, a, b):\n                ...\n    ```\n    ```py\n    RuntimeError: Unable to find any parametrizes in decorators, please rewrite decorator name to match any of detected names @{\'parametrize\'}  \n    ```\n\n- ### Parametrized method can be ran from command line only via pytest:\n    `$ cat test.py`\n    ```py\n    import unittest\n    from parametrize import parametrize\n    \n    class TestSomething(unittest.TestCase):\n        @parametrize(\'a\', (1, 2))\n        def test_something(self, a):\n            self.assertIsInstance(a, int)\n    ```\n    :white_check_mark: OK:\n    \n    `$ pytest test.py::TestSomething::test_something -v`\n    ```py\n     ...    \n     test.py::TestSomething.test_something[1] ✓                                       50% █████     \n     test.py::TestSomething.test_something[2] ✓                                      100% ██████████\n    \n    Results (0.07s):\n           2 passed\n    ```\n    :x: Won\'t work:\n    \n    `$ python -m unittest test.TestSomething.test_something`\n    ```py\n    Traceback (most recent call last):\n      ...\n    TypeError: don\'t know how to make test from: test_something[...]\n    ```\n- ### `@parametrize` cannot be used in interactive environments like REPL (It works in IPython though)\n- ### `@parametrize` cannot be used in cythonized code  \n',
    'author': 'MrMrRobat',
    'author_email': 'appkiller16@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MrMrRobat/parametrize/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
